# MIT License
# Copyright 2018 Ryan Hausen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ==============================================================================

"""Tests the model functions"""
import collections

import pytest
import numpy as np
import tensorflow as tf

import morpheus.core.base_model as base_model
import morpheus.core.unet as unet


class TestAssistant:
    """Makes things that the tests want."""

    @staticmethod
    def mock_dataset() -> collections.namedtuple:
        """Makes a compatible mock dataset.

        This snippet is adapted from morpheus.core.model.Morpheus
        """
        MockDataset = collections.namedtuple("Dataset", ["num_labels", "train", "test"])

        train_data = (TestAssistant.zero_array(), TestAssistant.zero_array())
        return MockDataset(5, train=train_data, test=train_data)

    @staticmethod
    def mock_hparams(
        inference: bool, batch_norm: bool, drop_out: bool, num_contractions: int
    ):
        """Makes a simple model config for testing.

        Args:
            inference (bool): boolean flag for inference
            batch_norm (bool): boolean flag for batch normalization in graph
            drop_out (bool): boolean flag for drop out in graph
            num_contractions (int): number of down/up sample pairs
        """

        num_filters = [1 for _ in range(num_contractions)]

        hparams = tf.contrib.training.HParams(
            inference=inference,
            num_epochs=1,
            batch_norm=batch_norm,
            drop_out=drop_out,
            down_filters=num_filters,
            up_filters=num_filters,
            learning_rate=0.1,
            dropout_rate=0.5,
            num_down_convs=1,
            num_up_convs=1,
            num_intermediate_filters=1,
        )

        return hparams

    @staticmethod
    def zero_tensor() -> tf.Tensor:
        """Makes a zero filled tensor with shape [3, 3]"""

        return tf.zeros([3, 3])

    @staticmethod
    def in_tensor(data_format: str) -> tf.Tensor:
        """Makes a sample tensor input tensor for shape testing.

        Args:
            data_format (str): 'channels_first' or 'channels_last'

        Returns:
            A tensor with shape [100, 80, 80, 5] if 'channels_last' or
            [100, 5, 80, 80] if 'channels_first'
        """

        shape = [100, 80, 80, 80]
        if data_format == "channels_first":
            shape[1] = 4
        else:
            shape[3] = 4

        return tf.zeros(shape, dtype=tf.float32)

    @staticmethod
    def zero_array() -> np.ndarray:
        """Makes a zero filled tensor with shape [3, 3]"""

        return np.zeros([3, 3])


@pytest.mark.unit
class TestBaseModel:
    """A class that tests the functions of morpheus.core.base_model.Model."""

    @staticmethod
    def test_model_fn_raises():
        """Tests that the non overridden model_fn raises NotImplemented."""

        model = base_model.Model(TestAssistant.mock_dataset())
        with pytest.raises(NotImplementedError):
            model.model_fn(TestAssistant.zero_tensor(), True)

    @staticmethod
    def test_build_graph_raises():
        """Tests that the non overridden model_fn raises on a build_graph call."""

        model = base_model.Model(TestAssistant.mock_dataset())
        with pytest.raises(NotImplementedError):
            model.build_graph(TestAssistant.zero_tensor(), True)

    @staticmethod
    def test_build_graph():
        """Tests the build_graph method."""

        model = base_model.Model(TestAssistant.mock_dataset())
        model.model_fn = lambda x, y: (x, y)
        expected_x = TestAssistant.zero_array()
        expected_is_training = True

        x, is_training = model.build_graph(expected_x, expected_is_training)

        np.testing.assert_array_equal(x, expected_x)
        assert is_training == expected_is_training

    @staticmethod
    def test_build_graph_singleton():
        """Tests the build_graph method, two calls."""

        model = base_model.Model(TestAssistant.mock_dataset())
        model.model_fn = lambda x, y: (x, y)
        expected_x = TestAssistant.zero_array()
        expected_is_training = True

        model.build_graph(expected_x, expected_is_training)
        x, is_training = model.build_graph(expected_x, expected_is_training)

        np.testing.assert_array_equal(x, expected_x)
        assert is_training == expected_is_training

    @staticmethod
    def test_train():
        """Test the train() method, doesn't raise."""
        model = base_model.Model(TestAssistant.mock_dataset())
        model.model_fn = lambda x, y: x
        model.train()

    @staticmethod
    def test_test():
        """Test the test() method, doesn't raise."""
        model = base_model.Model(TestAssistant.mock_dataset())
        model.model_fn = lambda x, y: x
        model.test()


@pytest.mark.unit
class TestUNet:
    """A class that tests the functions of morpheus.core.unet.Model"""

    @staticmethod
    def test_upsample_shape_doubles():
        inference, batch_norm, drop_out = True, True, True
        num_contractions = 1
        data_format = "channels_last"

        dataset = TestAssistant.mock_dataset()
        hparams = TestAssistant.mock_hparams(
            inference, batch_norm, drop_out, num_contractions
        )

        x = TestAssistant.in_tensor(data_format)

        expected_shape = x.shape.as_list()
        expected_shape[1] *= 2
        expected_shape[2] *= 2

        model = unet.Model(hparams, dataset, data_format)

        x = model.up_sample(x)

        actual_shape = x.shape.as_list()

        assert expected_shape == actual_shape

    @staticmethod
    def test_upsample_with_transpose_shape_doubles():
        inference, batch_norm, drop_out = True, True, True
        num_contractions = 1
        data_format = "channels_first"

        dataset = TestAssistant.mock_dataset()
        hparams = TestAssistant.mock_hparams(
            inference, batch_norm, drop_out, num_contractions
        )

        x = TestAssistant.in_tensor(data_format)

        expected_shape = x.shape.as_list()
        expected_shape[2] *= 2
        expected_shape[3] *= 2

        model = unet.Model(hparams, dataset, data_format)

        x = model.up_sample(x)

        actual_shape = x.shape.as_list()

        assert expected_shape == actual_shape

    @staticmethod
    def test_downsample_shape_halves():
        inference, batch_norm, drop_out = True, True, True
        num_contractions = 1
        data_format = "channels_last"

        dataset = TestAssistant.mock_dataset()
        hparams = TestAssistant.mock_hparams(
            inference, batch_norm, drop_out, num_contractions
        )

        x = TestAssistant.in_tensor(data_format)

        expected_shape = x.shape.as_list()
        expected_shape[1] /= 2
        expected_shape[2] /= 2

        model = unet.Model(hparams, dataset, data_format)

        x = model.down_sample(x)

        actual_shape = x.shape.as_list()

        assert expected_shape == actual_shape



# MIT License
# Copyright 2018 Ryan Hausen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ==============================================================================
"""Contains model code for Morpheus."""

import collections
import json
import os
from typing import List

import tensorflow as tf

import morpheus.core.unet


class Morpheus(morpheus.core.unet.Model):
    """The main class for the Morpheus model.

    This class takes an Tensorflow HParams object as an argument and it should 
    contain the following properties:

    Note if you are using a pretrained weights for inference only you need
    to mock the dataset object and use the default hparams.

    You can mock the dataset object calling Morpheus.mock_dataset().

    You can get the default HParams by calling Morpheus.inference_hparams().

    An example call for inference only
    
    >>> dataset = Morpheus.mock_dataset()
    >>> hparams = Morpheus.inference_hparams()
    >>> data_format = 'channels_last'
    >>> morph = Morpheus(hparams, dataset, data_format)

    Required HParams:
        * inference (bool): true if using pretrained model
        * down_filters (list): number of filters for each down conv section
        * num_down_convs (int): number of conv ops per down conv section
        * up_filters (list): number of filters for each up conv section
        * num_up_convs (int): number of conv ops per up conv section
        * batch_norm (bool): use batch normalization
        * dropout (bool): use dropout

    Optional HParams:
        * learning_rate (float): learning rate for training, required if inference is set to false
        * dropout_rate (float): the percentage of neurons to drop [0.0, 1.0]

    Args:
        hparams (tf.contrib.training.HParams): Model Hyperparameters
        dataset (tf.data.Dataset): dataset to use for training
        data_format: channels_first or channels_last

    TODO:
        * Make optimizer a paramter
    """

    def __init__(
        self,
        hparams: tf.contrib.training.HParams,
        dataset: tf.data.Dataset,
        data_format: str,
    ):
        """Inits Morpheus with hparams, dataset, data_format."""

        super().__init__(hparams, dataset, data_format)
        if not hparams.inference:

            self.opt = tf.train.AdamOptimizer(hparams.learning_rate)

    def loss_func(self, logits: tf.Tensor, labels: tf.Tensor) -> tf.Tensor:
        """Defines the loss function used in training.

        The loss function is defined by combining cross entropy loss calculated
        against all 5 classes and dice loss calculated against just the
        background class.

        Args:
            logits (tf.Tensor): output tensor from graph should be
                                [batch_size, width, height, 5]
            labels (tf.Tensor): labels used in training should be
                                [batch_size, width, height, 5]

        Returns:
            tf.Tensor: Tensor representing loss function.
        """

        flat_logits = tf.reshape(logits, [-1, 5])
        flat_y = tf.reshape(labels, [-1, 5])

        # Calculate weighted crossentropy ======================================
        # This is normally calculated by taking a count of the pixels assigned
        # to each class, but because we have continous values for each class
        # we sum the probabilities for each class in the pixels instead.
        xentropy_loss = tf.nn.softmax_cross_entropy_with_logits_v2(
            logits=flat_logits, labels=flat_y
        )

        dominant_class = tf.argmax(flat_y, axis=1, output_type=tf.int32)
        p_dominant_class = tf.reduce_max(flat_y, axis=1)

        class_coefficient = tf.zeros_like(xentropy_loss)
        for output_class_idx in range(5):
            class_pixels = tf.cast(
                tf.equal(output_class_idx, dominant_class), tf.float32
            )

            coef = tf.reduce_mean(class_pixels * p_dominant_class)
            class_coefficient = tf.add(class_coefficient, coef * class_pixels)

        class_coefficient = 1 / class_coefficient

        weighted_xentropy_loss = tf.reduce_mean(xentropy_loss * class_coefficient)
        # Calculate weighted crossentropy ======================================

        # Calculate dice loss ==================================================
        if self.data_format == "channels_first":
            yh_background = tf.nn.sigmoid(logits[:, -1, :, :])
            y_background = labels[:, -1, :, :]
        else:
            yh_background = tf.nn.sigmoid(logits[:, :, :, -1])
            y_background = labels[:, :, :, -1]

        dice_numerator = tf.reduce_sum(y_background * yh_background, axis=[1, 2])
        dice_denominator = tf.reduce_sum(y_background + yh_background, axis=[1, 2])

        dice_loss = tf.reduce_mean(2 * dice_numerator / dice_denominator)
        # Calculate dice loss ==================================================

        total_loss = weighted_xentropy_loss
        total_loss = total_loss + (1 - dice_loss)

        return total_loss

    def optimizer(self, loss: tf.Tensor) -> tf.Tensor:
        """Overrides the optimizer func in morpheus.core.unet

        Args:
            loss (tf.Tensor): The loss function tensor to pass to the optimizer

        Returns:
            tf.Tensor: the Tensor result of optimizer.minimize()
        """
        update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
        with tf.control_dependencies(update_ops):
            optimize = self.opt.minimize(loss)

        return optimize

    def train_metrics(
        self, logits: tf.Tensor, labels: tf.Tensor
    ) -> (
        (List[str], List[tf.Tensor]),
        List[tf.Tensor],
    ):  # overrides base method pylint: disable=no-self-use
        """Overrides the train_metrics func in morpheus.core.unet

        Args:
            logits (tf.Tensor): the output logits from the model
            labels (tf.Tensor): the labels used during training

        Returns:
            Tuple(Tuple(
                    List(str): names of metrics,
                    List(tf.Tensor): tensors for metrics
                  ),
                  List(tf.Tensor): Tensors for updating  running metrics
        """
        with tf.name_scope("train_metrics"):
            metrics_dict = Morpheus.eval_metrics(logits, labels)

        names, finalize, running = [], [], []

        for key in sorted(metrics_dict):
            names.append(key)
            finalize.append(metrics_dict[key][0])
            running.append(metrics_dict[key][1])

        return ([names, finalize], running)

    def test_metrics(
        self, logits: tf.Tensor, labels: tf.Tensor
    ) -> (
        (List[str], List[tf.Tensor]),
        List[tf.Tensor],
    ):  # overrides base method pylint: disable=no-self-use
        """Overrides the test_metrics func in morpheus.core.unet

        Args:
            logits (tf.Tensor): the output logits from the model
            labels (tf.Tensor): the labels used during training

        Returns:
            Tuple(Tuple(
                    List(str): names of metrics,
                    List(tf.Tensor): tensors for metrics
                  ),
                  List(tf.Tensor): Tensors for updating  running metrics
        """
        with tf.name_scope("test_metrics"):
            metrics_dict = Morpheus.eval_metrics(logits, labels)

        names, finalize, running = [], [], []

        for key in sorted(metrics_dict):
            names.append(key)
            finalize.append(metrics_dict[key][0])
            running.append(metrics_dict[key][1])

        return ([names, finalize], running)

    def inference(self, inputs: tf.Tensor) -> tf.Tensor:
        """Performs inference on input.

        Args:
            inputs (tf.Tensor): input tensor with shape
                                [batch_size, width, height, 5]

        Returns:
            A tf.Tensor of [batch_size, width, height, 5] representing the
            output the model, includes applying the softmax function.
        """

        return tf.nn.softmax(self.build_graph(inputs, False))

    @staticmethod
    def eval_metrics(yh: tf.Tensor, y: tf.Tensor) -> dict:
        """Function to generate metrics for evaluation during training.

        Args:
            yh (tf.Tensor): network output [n,h,w,c]
            y (tf.Tensor):  labels         [n,h,w,c]

        Returns:
            A dictionary collection of (tf.Tensor, tf.Tensor), where the keys
            are the names of the metrics and the values are running metric
            pairs. More infor on running accuracy metircs here:
            https://www.tensorflow.org/api_docs/python/tf/metrics/accuracy
        """
        metrics_dict = {}

        thresholds = [0.5, 0.6, 0.7, 0.8, 0.9]
        classes = ["spheroid", "disk", "irregular", "point_source", "background"]

        yh_bkg = tf.reshape(tf.nn.sigmoid(yh[:, :, :, -1]), [-1])
        y_bkg = tf.reshape(y[:, :, :, -1], [-1])
        for threshold in thresholds:
            name = "iou-{}".format(threshold)
            with tf.name_scope(name):
                preds = tf.cast(tf.greater_equal(yh_bkg, threshold), tf.int32)
                metric, update_op = tf.metrics.mean_iou(y_bkg, preds, 2, name=name)
                metrics_dict[name] = (metric, update_op)

        # Calculate the accuracy per class per pixel
        y = tf.reshape(y, [-1, 5])
        yh = tf.reshape(yh, [-1, 5])
        lbls = tf.argmax(y, 1)
        preds = tf.argmax(yh, 1)

        name = "overall"
        metric, update_op = tf.metrics.accuracy(lbls, preds, name=name)

        metrics_dict[name] = (metric, update_op)
        for i, _ in enumerate(classes):
            in_c = tf.equal(lbls, i)
            name = classes[i]
            metric, update_op = tf.metrics.accuracy(
                lbls, preds, weights=in_c, name=name
            )
            metrics_dict[name] = (metric, update_op)

        return metrics_dict

    @staticmethod
    def mock_dataset() -> collections.namedtuple:
        """Generates a mockdataset for inference.

        Returns:
            A collections.namedtuple object that can be passed in place of a
            tf.data.Dataset for 'dataset' argument in the constructor
        """
        MockDataset = collections.namedtuple("Dataset", ["num_labels"])
        return MockDataset(5)

    @staticmethod
    def inference_hparams() -> tf.contrib.training.HParams:
        """Generates a mockdataset for inference.

        Returns:
            a tf.contrib.training.HParams object with the settings for inference
        """
        config_path = os.path.join(os.path.dirname(__file__), "model_config.json")

        with open(config_path, "r") as f:
            return tf.contrib.training.HParams(**json.load(f))

    @staticmethod
    def get_weights_dir() -> str:
        """Returns the location of the weights for tf.Saver."""
        return os.path.join(os.path.dirname(__file__), "model_weights")



# MIT License
# Copyright 2018 Ryan Hausen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ==============================================================================
"""Helper classes used in Morpheus."""

import os

from types import FunctionType
from typing import List
from typing import Iterable
from typing import Tuple

import numpy as np
import tensorflow as tf

from colorama import init, Fore
from astropy.io import fits

init(autoreset=True)


class TFLogger:
    """A helper class to color the logging text in Tensorflow."""

    RED = lambda s: Fore.RED + str(s) + Fore.RESET
    BLUE = lambda s: Fore.BLUE + str(s) + Fore.RESET
    YELLOW = lambda s: Fore.YELLOW + str(s) + Fore.RESET
    GREEN = lambda s: Fore.GREEN + str(s) + Fore.RESET
    LIGHTRED = lambda s: Fore.LIGHTRED_EX + str(s) + Fore.RESET

    @staticmethod
    def info(msg: str) -> None:
        """Log at info level in green.

    @staticmethod
    @staticmethodgged
    @staticmethod
        Returns:
            None
        """
        tf.logging.info(TFLogger.GREEN(msg))

    @staticmethod
    def debug(msg: str) -> None:
        """Log at debug level in yellow.

        Args:
            msg (str): The string to be logged

        Returns:
            None
        """
        tf.logging.debug(TFLogger.YELLOW(msg))

    @staticmethod
    def warn(msg: str) -> None:
        """Log at warn level in lightred.

        Args:
            msg (str): The string to be logged

        Returns:
            None
        """
        tf.logging.warning(TFLogger.LIGHTRED(msg))

    @staticmethod
    def error(msg: str):
        """Log at error level in red.

        Args:
            msg (str): The string to be logged

        Returns:
            None
        """
        tf.logging.error(TFLogger.RED(msg))

    @staticmethod
    def tensor_shape(tensor: tf.Tensor, log_func=None, format_str="[{}]::{}") -> None:
        """Log the the shape of tensor 't'.

        Args:
            tensor (tf.Tensor): A tensorflow Tensor
            logging_func (func): logging function to to use, default
                                tf_logger.debug
            format_str (str): A string that will be passed will have .format called
                            on it and given two arguments in the following order:
                            - tensor_name
                            - tensor_shape
        Returns:
            None
        """
        if log_func is None:
            log_func = TFLogger.debug

        log_func(format_str.format(tensor.name, tensor.shape.as_list()))


class OptionalFunc:
    """Descriptor protocol for functions that don't have to overriden.

    This is a helper class that is used to stub methods that don't have to
    be overridden.
    """

    def placeholder(self, *args):
        """Placeholder function used as default in __init__"""
        return list(args)

    def __init__(self, warn_msg: str, init_func: FunctionType = None):
        """"""
        self._warn_msg = warn_msg
        self._func = init_func if init_func else self.placeholder
        self._is_default = True

    def __get__(
        self, obj, type=None  # pylint: disable=redefined-builtin
    ) -> FunctionType:
        if self._is_default:
            TFLogger.warn(self._warn_msg)

        return self._func

    def __set__(self, obj, value) -> None:
        self._is_default = False
        self._func = value


class FitsHelper:
    """A class that handles basic FITS file functions."""

    @staticmethod
    def create_file(file_name: str, data_shape: tuple, dtype) -> None:
        """Creates a fits file without loading it into memory.

        This is a helper method to create large FITS files without loading an
        array into memory. The method follows the direction given at:
        http://docs.astropy.org/en/stable/generated/examples/io/skip_create-large-fits.html


        Args:
            file_name (str): the complete path to the file to be created.
            data_shape (tuple): a tuple describe the shape of the file to be
                                created
            dtype (numpy datatype): the numpy datatype used in the array

        Raises:
            ValueError if dtype is not one of:
                - np.unit8
                - np.int16
                - np.int32
                - np.float32
                - np.float64


        TODO: Figure out why this throws warning about size occasionally
              when files that are created by it are opened
        """
        bytes_per_value = 0

        if dtype == np.uint8:
            bytes_per_value = 1
        elif dtype == np.int16:
            bytes_per_value = 2
        elif dtype == np.int32:
            bytes_per_value = 4
        elif dtype == np.float32:
            bytes_per_value = 4
        elif dtype == np.float64:
            bytes_per_value = 8

        if bytes_per_value == 0:
            raise ValueError("Invalid dtype")

        stub_size = [100, 100]
        if len(data_shape) == 3:
            stub_size.append(5)
        stub = np.zeros(stub_size, dtype=dtype)

        hdu = fits.PrimaryHDU(data=stub)
        header = hdu.header
        while len(header) < (36 * 4 - 1):
            header.append()

        header["NAXIS1"] = data_shape[1]
        header["NAXIS2"] = data_shape[0]
        if len(data_shape) == 3:
            header["NAXIS3"] = data_shape[2]

        header.tofile(file_name)

        with open(file_name, "rb+") as f:
            header_size = len(header.tostring())
            data_size = (np.prod(data_shape) * bytes_per_value) - 1

            f.seek(header_size + data_size)
            f.write(b"\0")

    @staticmethod
    def get_files(
        file_names: List[str], mode: str = "readonly"
    ) -> (List[fits.HDUList], List[np.ndarray]):
        """Gets the HDULS and data handles for all the files in file_names.

        This is a convience function to opening multiple FITS files using
        memmap.

        Args:
            file_names (List[str]): a list of file names including paths to FITS
                                    files
            mode (str): the mode to pass to fits.open

        Returns:
            Tuple of a list numpy arrays that are the mmapped data handles for
            each of the FITS files and the HDULs that go along with them
        """
        arrays = []
        hduls = []

        for f in file_names:
            hdul = fits.open(f, mode=mode, memmap=True)
            arrays.append(hdul[0].data)  # Astropy problem pylint: disable=E1101
            hduls.append(hdul)

        return hduls, arrays

    @staticmethod
    def create_mean_var_files(
        shape: List[int], out_dir: str
    ) -> (List[fits.HDUList], List[np.ndarray]):
        """Creates the output fits files for the mean/variance morpheus output.

            Args:
                shape (List[int]): The shape to use when making the FITS files
                out_dir (str): the directory to place the files in. Will make it
                               if it doesn't already exist.

            Returns:
                List[fits.HDUList]: for the created files
                Dict(str, np.ndarray): a dictionary where the key is the data
                                       descriptor and the value is the memmapped
                                       data numpy array
        """

        data_keys = []
        file_names = []
        for morph in LabelHelper.MORPHOLOGIES:
            for t in ["mean", "var"]:
                f = os.path.join(out_dir, f"{morph}_{t}.fits")
                file_names.append(f)
                data_keys.append(f"{morph}_{t}")

                FitsHelper.create_file(f, shape, np.float32)

        hduls, arrays = FitsHelper.get_files(file_names, mode="update")

        return hduls, {k: v for k, v in zip(data_keys, arrays)}

    @staticmethod
    def create_rank_vote_files(
        shape: List[int], out_dir: str
    ) -> (List[fits.HDUList], List[np.ndarray]):
        """Creates the output fits files for the rank vote morpheus output.

            Args:
                shape (List[int]): The shape to use when making the FITS files
                out_dir (str): the directory to place the files in. Will make it
                               if it doesn't already exist.

            Returns:
                List[fits.HDUList]: for the created files
                Dict(str, np.ndarray): a dictionary where the key is the data
                                       descriptor and the value is the memmapped
                                       data numpy array
        """

        data_keys = []
        file_names = []
        for morph in LabelHelper.MORPHOLOGIES:
            f = os.path.join(out_dir, f"{morph}.fits")
            file_names.append(f)
            data_keys.append(morph)

            FitsHelper.create_file(f, shape, np.float32)

        hduls, arrays = FitsHelper.get_files(file_names, mode="update")

        return hduls, {k: v for k, v in zip(data_keys, arrays)}

    @staticmethod
    def create_n_file(
        shape: List[int], out_dir: str
    ) -> (List[fits.HDUList], List[np.ndarray]):
        """Creates the output fits files for the rank vote morpheus output.

            Args:
                shape (List[int]): The shape to use when making the FITS files
                out_dir (str): the directory to place the files in. Will make it
                               if it doesn't already exist.

            Returns:
                List[fits.HDUList]: for the created files
                Dict(str, np.ndarray): a dictionary where the key is the data
                                       descriptor and the value is the memmapped
                                       data numpy array
        """

        n_path = os.path.join(out_dir, "n.fits")
        FitsHelper.create_file(n_path, shape, np.int16)
        hduls, arrays = FitsHelper.get_files([n_path], mode="update")

        return hduls, {"n": arrays[0]}


class LabelHelper:
    """Class to help with label updates.

    Class Variables:
    UPDATE_MASK (np.ndarray): the (40, 40) integer array that indicates which
                              parts of the output of the model to include in the
                              calculations. default: innermost (30,30)
    UPDATE_MASK_N (np.ndarray): the (40, 40) integer array that indicates which
                                parts of the count 'n' to udpate. default:
                                all (40, 40)
    """

    MORPHOLOGIES = ["spheroid", "disk", "irregular", "point_source", "background"]

    UPDATE_MASK = np.pad(np.ones([30, 30]), 5, mode="constant").astype(np.int16)
    UPDATE_MASK_N = np.ones([40, 40], dtype=np.int16)

    @staticmethod
    def index_generator(dim0: int, dim1: int) -> Iterable[Tuple[int, int]]:
        """Creates a generator that returns indicies to iterate over a 2d array.

        Args:
            dim0 (int): The upper limit to iterate upto for the first dimension
            dim1 (int): The upper limit to iterate upto for the second dimension

        Returns:
            A generator that yields indicies to iterate over a 2d array with
            shape [dim0, dim1]
        """
        for y in range(dim0):
            for x in range(dim1):
                yield (y, x)

    @staticmethod
    def windowed_index_generator(dim0: int, dim1: int) -> Iterable[Tuple[int, int]]:
        """Creates a generator that returns window limited indicies over a 2d array.

        THe generator returned by this method will yield the indicies for the use
        of a sliding window of size `N_UPDATE_MASK.shape` over a 2d array with
        the size `(dim0, dim1)`.

        Args:
            dim0 (int): The upper limit to iterate upto for the first dimension
            dim1 (int): The upper limit to iterate upto for the second dimension

        Returns:
            A generator that yields indicies to iterate over a 2d array with
            shape [dim0, dim1]
        """

        window_y, window_x = LabelHelper.UPDATE_MASK_N.shape
        final_y = dim0 - window_y + 1
        final_x = dim1 - window_x + 1

        return LabelHelper.index_generator(final_y, final_x)

    @staticmethod
    def get_final_map(shape: List[int], y: int, x: int):
        """Creates a pixel mapping that flags pixels that won't be updated again.

        Args:
            shape (List[int]): the shape of the array that x and y are indexing
            y (int): the current y index
            x (int): the current x index

        Returns:
            A list of relative indicies that won't be updated again.
        """
        final_map = []

        end_y = y == (shape[0] - LabelHelper.UPDATE_MASK_N.shape[0])
        end_x = x == (shape[1] - LabelHelper.UPDATE_MASK_N.shape[1])

        if end_y and end_x:
            for _y in range(5, 35):
                for _x in range(5, 35):
                    final_map.append((_y, _x))
        else:
            if end_x:
                final_map.extend([(5, _x) for _x in range(5, 35)])
            if end_y:
                final_map.extend([(_y, 5) for _y in range(5, 35)])

        if not final_map:
            final_map.append((5, 5))

        return final_map

    @staticmethod
    def iterative_mean(
        n: np.ndarray, curr_mean: np.ndarray, x_n: np.ndarray, update_mask: np.ndarray
    ):
        """Calculates the mean of collection in an online fashion.

        The values are calculated using the following equation:
        http://people.ds.cam.ac.uk/fanf2/hermes/doc/antiforgery/stats.pdf, eq. 4

        Args:
            n (np.ndarray): a 2d array containg the number of terms in mean so
                            far,
            prev_mean (np.ndarray): the current calculated mean.
            x_n (np.ndarray): the new values to add to the mean
            update_mask (np.ndarray): a 2d boolean array indicating which
                                      indicies in the array should be updated.

        Returns:
            An array with the same shape as the curr_mean with the newly
            calculated mean values.
        """
        n[n == 0] = 1
        return curr_mean + ((x_n - curr_mean) / n * update_mask)

    @staticmethod
    def iterative_variance(
        prev_sn: np.ndarray,
        x_n: np.ndarray,
        prev_mean: np.ndarray,
        curr_mean: np.ndarray,
        update_mask: np.ndarray,
    ):
        """The first of two methods used to calculate the variance online.

        This method specifically calculates the $S_n$ value as indicated in
        equation 24 from:

        http://people.ds.cam.ac.uk/fanf2/hermes/doc/antiforgery/stats.pdf

        Args:
            prev_sn (np.ndarray): the $S_n$ value from the previous step
            x_n (np.ndarray): the current incoming values
            prev_mean (np.ndarray): the mean that was previously calculated
            curr_mean (np.ndarray): the mean, including the current values
            update_mask (np.ndarray): a boolean mask indicating which values to
                                      update

        Returns:
            An np.ndarray containg the current value for $S_n$


        """
        return prev_sn + ((x_n - prev_mean) * (x_n - curr_mean) * update_mask)

    @staticmethod
    def finalize_variance(
        n: np.ndarray, curr_sn: np.ndarray, final_map: List[Tuple[int, int]]
    ):
        """The second of two methods used to calculate the variance online.

        This method calcaulates the final variance value using equation 25 from

        http://people.ds.cam.ac.uk/fanf2/hermes/doc/antiforgery/stats.pdf

        but without performing the square root.

        Args:
            n (np.ndarray): the current number of values included in the calculation
            curr_sn (np.ndarray): the current $S_n$ values
            final_map List[(y, x)]: a list of indicies to calculate the final
                                    variance for

        Returns:
            A np.ndarray with the current $S_n$ values and variance values for
            all indicies in final_map
        """
        final_n = np.ones_like(n)
        for y, x in final_map:
            final_n[y, x] = n[y, x]

        return curr_sn / final_n

    @staticmethod
    def iterative_rank_vote(
        x_n: np.ndarray, prev_count: np.ndarray, update_mask: np.ndarray
    ):
        """Calculates the updated values for the rank vote labels for a one class.

        Args:
            x_n (np.ndarray): the current rank vote values for the class being
                              updated
            prev_count (np.ndarray): the array containing the running totals,
                                     should be shaped as [labels, height, width]
            update_mask (np.ndarray): a boolean array indicating which values to
                                      update

        Returns:
            A numpy array containing the updated count values
        """
        update = np.zeros_like(prev_count)

        for i in range(update.shape[0]):
            for j in range(update.shape[1]):
                if update_mask[i, j] and (x_n[i, j] == 4):
                    update[i, j] = 1

        count = prev_count + update

        return count

    @staticmethod
    def update_ns(data: dict, batch_idx: List[Tuple[int, int]], inc: int = 1) -> None:
        """Updates the n values by `inc`.

        Args:
            data (dict): a dictionary of numpy arrays containing the data
            batch_idx (List[Tuple[int, int]]): a list of indicies to update
            inc (int): the number to increment `n` by. Default=1

        Returns
            None
        """
        window_y, window_x = LabelHelper.UPDATE_MASK_N.shape
        for y, x in batch_idx:
            ys = slice(y, y + window_y)
            xs = slice(x, x + window_x)

            ns = data["n"][ys, xs]
            n_update = LabelHelper.UPDATE_MASK_N * LabelHelper.UPDATE_MASK * inc
            ns = ns + n_update
            data["n"][ys, xs] = ns

    @staticmethod
    def update_mean_var(
        data: dict, labels: np.ndarray, batch_idx: List[Tuple[int, int]]
    ):
        """Updates the mean and variance outputs with the new model values.

        Args:
            data (dict): a dict of numpy arrays containing the data
            labels (np.ndarray): the new output from the model
            batch_idx (List[Tuple[int, int]]): a list of indicies to update

        Returns:
            None
        """

        window_y, window_x = LabelHelper.UPDATE_MASK_N.shape
        total_shape = data["n"].shape
        for i, l in enumerate(labels):
            y, x = batch_idx[i]
            ys = slice(y, y + window_y)
            xs = slice(x, x + window_x)

            final_map = LabelHelper.get_final_map(total_shape, y, x)
            n = data["n"][ys, xs]
            for j, morph in enumerate(LabelHelper.MORPHOLOGIES):
                k_mean = f"{morph}_mean"
                k_var = f"{morph}_var"

                x_n = l[:, :, j]
                prev_mean = data[k_mean][ys, xs]
                prev_var = data[k_var][ys, xs]

                mean = LabelHelper.iterative_mean(
                    n, prev_mean, x_n, LabelHelper.UPDATE_MASK
                )

                var = LabelHelper.iterative_variance(
                    prev_var, x_n, prev_mean, mean, LabelHelper.UPDATE_MASK
                )
                var = LabelHelper.finalize_variance(n, var, final_map)

                data[k_mean][ys, xs] = mean
                data[k_var][ys, xs] = var

    @staticmethod
    def update_rank_vote(
        data: dict, labels: np.ndarray, batch_idx: List[Tuple[int, int]]
    ) -> None:
        """Updates the rank vote values with the new output.

        Args:
            data (dict): data (dict): a dict of numpy arrays containing the data
            labels (np.ndarray): the new output from the model
            batch_idx (List[Tuple[int, int]]): a list of indicies to update

        Returns:
            None
        """

        window_y, window_x = LabelHelper.UPDATE_MASK_N.shape
        for i, l in enumerate(labels):
            y, x = batch_idx[i]
            ys = slice(y, y + window_y)
            xs = slice(x, x + window_x)

            ranked = l.argsort().argsort()
            for j, morph in enumerate(LabelHelper.MORPHOLOGIES):
                prev_count = data[morph][ys, xs]

                count = LabelHelper.iterative_rank_vote(
                    ranked[:, :, j], prev_count, LabelHelper.UPDATE_MASK
                )

                data[morph][ys, xs] = count

    @staticmethod
    def update_labels(
        data: dict, labels: np.ndarray, batch_idx: List[Tuple[int, int]], out_type: str
    ) -> None:
        """Updates the running total label values with the new output values.

        Args:
            data (dict): data (dict): a dict of numpy arrays containing the data
            labels (np.ndarray): the new output from the model
            batch_idx (List[Tuple[int, int]]): a list of indicies to update
            out_type (str): indicates which type of output to update must be
                            one of ['mean_var', 'rank_vote', 'both']

        Returns:
            None
        """

        LabelHelper.update_ns(data, batch_idx)

        if out_type in ["both", "mean_var"]:
            LabelHelper.update_mean_var(data, labels, batch_idx)
        if out_type in ["both", "rank_vote"]:
            LabelHelper.update_rank_vote(data, labels, batch_idx)

    @staticmethod
    def make_mean_var_arrays(shape: Tuple[int, int]) -> dict:
        """Create output arrays for use in in-memory classification.

        Args:
            shape (Tuple[int]): The 2d (width, height) for to create the arrays

        Returns
            A dictionary with keys being the arrays description and values being
            the array itself
        """

        arrays = {}

        for morph in LabelHelper.MORPHOLOGIES:
            for t in ["mean", "var"]:
                arrays[f"{morph}_{t}"] = np.zeros(shape, dtype=np.float32)

        return arrays

    @staticmethod
    def make_rank_vote_arrays(shape: Tuple[int, int]) -> dict:
        """Create output arrays for use in in-memory classification.

        Args:
            shape (Tuple[int]): The 2d (width, height) for to create the arrays

        Returns
            A dictionary with keys being the arrays description and values being
            the array itself
        """
        shape = [shape[0], shape[1]]
        arrays = {}

        for morph in LabelHelper.MORPHOLOGIES:
            arrays[morph] = np.zeros(shape, dtype=np.float32)

        return arrays

    @staticmethod
    def make_n_array(shape: Tuple[int, int]) -> dict:
        """Create output array for use in in-memory classification.

        Args:
            shape (Tuple[int]): The 2d (width, height) for to create the arrays

        Returns
            A dictionary with keys being the arrays description and values being
            the array itself
        """
        return {"n": np.zeros(shape, dtype=np.int16)}

    @staticmethod
    def finalize_rank_vote(data: dict) -> None:
        """Finalize the rank vote by dividing by n.

        Args:
            data (dict): a dict of numpy arrays containing the data

        TODO: Refactor to accomodate large files

        Returns:
            None
        """

        n = data["n"]

        for morph in LabelHelper.MORPHOLOGIES:
            np.divide(data[morph], n, out=data[morph], where=n != 0)



# MIT License
# Copyright 2018 Ryan Hausen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ==============================================================================
"""An interface for interacting with Morpheus"""
import os
import time
from subprocess import Popen
from typing import Iterable, List, Tuple, Callable, Dict

import imageio
import numpy as np
import tensorflow as tf
from astropy.io import fits
from matplotlib.colors import hsv_to_rgb
from scipy import ndimage as ndi
from skimage.feature import peak_local_max
from skimage.filters import sobel
from skimage.measure import regionprops
from skimage.morphology import watershed
from tqdm import tqdm

import morpheus.core.helpers as helpers
import morpheus.core.model as model


class Classifier:
    """Primary interface for the use of Morpheus.

    Images can be classified by calling
    :py:meth:`~morpheus.classifier.Classifier.classify_arrays` and passing
    numpy arrays or by calling
    :py:meth:`~morpheus.classifier.Classifier.classify_files` and passing string
    FITS file locations.

    After an image this this class offers some post processing functionality by
    generating segmentation maps using
    :py:meth:`~morpheus.classifier.Classifier.make_segmap` and colorized
    morphological classifications using
    :py:meth:`~morpheus.classifier.Classifier.colorize_rank_vote_output`

    For more examples see the `documentation <https://morpheus-astro.readthedocs.io/>`_.
    """

    __graph = None
    __session = None
    __X = tf.placeholder(tf.float32, shape=[None, 40, 40, 4])

    @staticmethod
    def catalog_arrays(
        h: str = None,
        j: str = None,
        z: str = None,
        v: str = None,
        out_dir: str = None,
        pad: bool = False,
        batch_size: int = 1000,
        out_type: str = "rank_vote",
        segmap_mask: np.ndarray = None,
        return_segmap: bool = False,
        return_morphological_map: bool = False,
    ) -> Dict:
        """Generate a source catalog using Morpheus.

        This method generates a catalog of the sources in the array. It is a
        convenience method that combines the other methods into a single action.
        The following operations are performed:

        1. The input image is classified using
           :py:meth:`~morpheus.classifier.Classifier.classify_arrays`
        2. The classification is used to generate a segmentation map using
           :py:meth:`~morpheus.classifier.Classifier.make_segmap`
        3. The the segmentation maps and the classification output are combined
           to create aggregate morphological classfifications for each source
           in the segmentation map


        Args:
            h (np.ndarray): the H band values for an image
            j (np.ndarray): the J band values for an image
            z (np.ndarray): the Z band values for an image
            v (np.ndarray): the V band values for an iamge
            out_dir (str): the location where to save the output files
                           if None returns the output in memory only.
            pad (bool): if True pad the input with zeros, so that every pixel is
                        classified the same number of times. If False, don't pad.
                        (Not implemented)
            batch_size (int): the number of image sections to process at a time
            out_type (str): how to process the output from Morpheus. If
                            'mean_var' record output using mean and variance, If
                            'rank_vote' record output as the normalized vote
                            count. If 'both' record both outputs.
            segmap_mask (np.ndarray): a boolean mask that indicates all valid
                                      pixels in the images. If None, all pixels
                                      are considered valid.
            return_segmap (bool): if `True` include the segmentation map in the
                                  returned dictionary with the key `segmentation_map`
            return_morphological_map (bool): if `True` include the morhological
                                             output maps for the input in the returned
                                             dictionary with the key `morphological_map`

        Returns:
            A dictionary containing the catalog of the classification output
            and the associated catalog. The dictionary contains the following
            keys:

            - 'classifier_output' The output from :py:meth:`~morpheus.classifier.Classifier.classify_arrays`
            - 'segmentation_map' The output from :py:meth:`~morpheus.classifier.Classifier.make_segmap`
            - 'catalog' A list of dictionary objects containing the catalog


        Raises:
            ValueError if out_type is not one of ['mean_var', 'rank_vote', 'both']
        """
        classified = Classifier.classify_arrays(
            h=h,
            j=j,
            v=v,
            z=z,
            out_dir=out_dir,
            pad=pad,
            batch_size=batch_size,
            out_type=out_type,
        )

        segmap = Classifier.make_segmap(
            data=classified, h=h, out_dir=out_dir, min_distance=20, mask=segmap_mask
        )

        # add catalog here
        catalog = Classifier.make_catalog(data=classified, flux=h, segmap=segmap)

        outs = {"catalog": catalog}

        if return_segmap:
            outs["segmentation_map"] = segmap
        if return_morphological_map:
            outs["morphological_map"] = classified

        return outs

    @staticmethod
    def classify_files(
        h: str = None,
        j: str = None,
        z: str = None,
        v: str = None,
        out_dir: str = ".",
        pad: bool = False,
        batch_size: int = 1000,
        out_type: str = "rank_vote",
        gpus: List[int] = None,
        parallel_check_interval: int = 15,
    ) -> None:
        """Classify FITS files using Morpheus.

        Args:
            h (str): The file location of the H band FITS file
            j (str): The file location of the J band FITS file
            z (str): The file location of the Z band FITS file
            v (str): The file location of the V band FITS file
            out_dir (str): The location where to save the output files
            pad (bool): if True pad the input with zeros, so that every pixel is
                        classified the same number of times. If False, don't pad.
                        (Not implemented yet)
            batch_size (int): the number of image sections to process at a time
            out_type (str): how to process the output from Morpheus. If
                            'mean_var' record output using mean and variance, If
                            'rank_vote' record output as the normalized vote
                            count. If 'both' record both outputs.
            gpus (List[int]): A list of the CUDA gpu ID's to use for a
                              parallel classification.
            parallel_check_interval (int): If gpus are given, then this is the number
                                           of minutes to wait between polling each
                                           subprocess for completion

        Returns:
            None
        """

        hduls, [h, j, v, z] = Classifier._parse_files(h, j, v, z)

        is_serial_run = (gpus is None) or (not gpus) or (len(gpus) == 1)
        if is_serial_run:
            Classifier.classify_arrays(
                h=h,
                j=j,
                v=v,
                z=z,
                out_type=out_type,
                pad=pad,
                out_dir=out_dir,
                batch_size=batch_size,
            )
        else:
            Classifier._build_parallel_classification_structure(
                [h, j, v, z], gpus, out_dir
            )
            Classifier._run_parallel_jobs(gpus, out_dir, parallel_check_interval)
            Classifier._stitch_parallel_classifications(out_dir)

        for hdul in hduls:
            hdul.close()

    @staticmethod
    def classify_arrays(
        h: np.ndarray = None,
        j: np.ndarray = None,
        z: np.ndarray = None,
        v: np.ndarray = None,
        out_dir: str = None,
        pad: bool = False,
        batch_size: int = 1000,
        out_type: str = "rank_vote",
    ) -> Dict:
        """Classify numpy arrays using Morpheus.

        Args:
            h (np.ndarray): the H band values for an image
            j (np.ndarray): the J band values for an image
            z (np.ndarray): the Z band values for an image
            v (np.ndarray): the V band values for an iamge
            out_dir (str): The location where to save the output files
                           if None returns the output in memory only.
            pad (bool): if True pad the input with zeros, so that every pixel is
                        classified the same number of times. If False, don't pad.
                        (Not implemented)
            batch_size (int): the number of image sections to process at a time
            out_type (str): how to process the output from Morpheus. If
                            'mean_var' record output using mean and variance, If
                            'rank_vote' record output as the normalized vote
                            count. If 'both' record both outputs.

        Returns:
            A dictionary containing the output classifications.

        Raises:
            ValueError if out_type is not one of ['mean_var', 'rank_vote', 'both']
        """
        Classifier._variables_not_none(["h", "j", "z", "v"], [h, j, z, v])
        Classifier._arrays_same_size([h, j, z, v])

        if out_type not in ["mean_var", "rank_vote", "both"]:
            raise ValueError("Invalid value for `out_type`")

        mean_var = out_type in ["mean_var", "both"]
        rank_vote = out_type in ["rank_vote", "both"]

        if pad:
            raise NotImplementedError("pad=True has not been implemented yet")

        shape = h.shape

        hduls = []
        data = {}
        if out_dir:
            if mean_var:
                hs, ds = helpers.FitsHelper.create_mean_var_files(shape, out_dir)
                hduls.extend(hs)
                data.update(ds)
            if rank_vote:
                hs, ds = helpers.FitsHelper.create_rank_vote_files(shape, out_dir)
                hduls.extend(hs)
                data.update(ds)

            hs, ds = helpers.FitsHelper.create_n_file(shape, out_dir)
            hduls.extend(hs)
            data.update(ds)
        else:
            if mean_var:
                data.update(helpers.LabelHelper.make_mean_var_arrays(shape))
            if rank_vote:
                data.update(helpers.LabelHelper.make_rank_vote_arrays(shape))

            data.update(helpers.LabelHelper.make_n_array(shape))

        indicies = helpers.LabelHelper.windowed_index_generator(*shape)

        window_y, window_x = helpers.LabelHelper.UPDATE_MASK_N.shape
        batch_estimate = shape[0] - window_y + 1
        batch_estimate *= shape[1] - window_x + 1
        batch_estimate = batch_estimate // batch_size
        pbar = tqdm(total=batch_estimate, desc="classifying", unit="batch")

        while True:
            batch = []
            batch_idx = []

            for _ in range(batch_size):
                try:
                    y, x = next(indicies)
                except StopIteration:
                    break

                combined = np.array(
                    [img[y : y + window_y, x : x + window_x] for img in [h, j, v, z]]
                )
                batch.append(Classifier._standardize_img(combined))
                batch_idx.append((y, x))

            if not batch:
                break

            batch = np.array(batch)

            labels = Classifier._call_morpheus(batch)

            helpers.LabelHelper.update_labels(data, labels, batch_idx, out_type)

            pbar.update()

        if rank_vote:
            helpers.LabelHelper.finalize_rank_vote(data)

        for hdul in hduls:
            hdul.close()

        return data

    @staticmethod
    def _standardize_img(img: np.ndarray) -> np.ndarray:
        """Standardizes an input img to mean 0 and unit variance.

        Uses the formula described in:

        https://www.tensorflow.org/api_docs/python/tf/image/per_image_standardization

        Args:
            img (np.ndarray): the input array to standardize

        Returns:
            The standardized input
        """
        num = img - img.mean()
        denom = max(img.std(), 1 / np.sqrt(np.prod(img.shape)))
        return num / denom

    @staticmethod
    def _arrays_same_size(arrays: List[np.ndarray]) -> None:
        """Verifies that all arrays are the same shape.

        Args:
            arrays (List[np.ndarray]): List of arrays that should have the same
                                       shape.

        Returns:
            None

        Raises:
            ValueError if arrays are not the same shape
        """

        arr_shapes = [a.shape for a in arrays]

        arr_comp = arr_shapes[0]
        arr_to_comp = arr_shapes[1:]

        if not np.array_equiv(arr_comp, arr_to_comp):
            raise ValueError(f"All shapes not the same: {arr_shapes}.")

    @staticmethod
    def _variables_not_none(names: List[str], values: List[np.ndarray]) -> None:
        """Verifies that all variables are not None.

        Args:
            names (List[str]): list of names of variables in the same order as
                               `values`
            names (List[np.ndarray]): list of numpy arrays that should not be
                                      None

        Returns:
            None

        Raises:
            ValueError if a variable is None

        """

        nones = []
        for name, value in zip(names, values):
            if value is None:
                nones.append(name)

        if nones:
            raise ValueError("{} should not be None".format(nones))

    @staticmethod
    def _parse_files(
        h: str, j: str, v: str, z: str
    ) -> Tuple[List[fits.HDUList], List[np.ndarray]]:
        """Validates that files exist. And returns the corresponding arrays.

        Args:
            h (str): the file location of the H band img
            j (str): the file location of the J band img
            v (str): the file location of the V band img
            z (str): the file location of the Z bnad img

        Returns:
            A tuple containing the a (List[HDUL], List[np.ndarray])

        Raises:
            ValueError if a variable is None

        """
        Classifier._variables_not_none(["h", "j", "z", "v"], [h, j, z, v])

        return helpers.FitsHelper.get_files([h, j, v, z])

    @staticmethod
    def _call_morpheus(batch: np.ndarray) -> np.ndarray:
        """Use morpheus to classify a batch of input values.

        Morpheus is called as a singleton using this method.

        Args:
            batch (np.ndarray): The input data in the shape
                                [batch, channels, width, height]

        Returns:
            The classified numpy array with shape [batch, width, height, channels]

        """
        batch = np.transpose(batch, axes=[0, 2, 3, 1])

        if Classifier.__graph is None:
            config = model.Morpheus.inference_hparams()
            inference_dataset = model.Morpheus.mock_dataset()

            # build graph
            m = model.Morpheus(config, inference_dataset, "channels_last")
            Classifier.__graph = m.inference(Classifier.__X)

            # get weights
            saver = tf.train.Saver()
            Classifier.__session = tf.Session()
            w_location = model.Morpheus.get_weights_dir()
            saver.restore(Classifier.__session, tf.train.latest_checkpoint(w_location))

        return Classifier.__session.run(
            Classifier.__graph, feed_dict={Classifier.__X: batch}
        )

    @staticmethod
    def _get_split_length(shape: List[int], num_gpus: int) -> int:
        """Calculate the size of the sub images for classification.

        Args:
            shape (List[int]): the shape of the array to be split
            num_gpus (int): the number of splits to make

        Returns:
            The length of each split along axis 0

        TODO: Implement splits along other axes
        """

        return (shape[0] + (num_gpus - 1) * 40) // num_gpus

    @staticmethod
    def _get_split_slice_generator(
        shape: Tuple[int], num_gpus: int, slice_length: int
    ) -> Iterable[slice]:
        """Creates a generator that yields `slice` objects to split imgs.

        Args:
            shape (Tuple[int]): The shape of the array to be split
            num_gpus (int): The number of splits to make
            split_length (int): The length each slice should be

        Returns
            A generator that yields slice objects

        TODO: Implement splits along other axes
        """

        idx = 0
        for i in range(num_gpus):
            start_idx = max(idx - 40, 0)

            if i == num_gpus - 1:
                end_idx = shape[0]
            else:
                end_idx = start_idx + slice_length

            idx = end_idx

            yield slice(start_idx, end_idx)

    @staticmethod
    def _make_runnable_file(path: str) -> None:
        """Creates a file at `path` that classfies local FITS files.

        Args:
            path (str): The dir to save the file in

        Returns:
            None
        """

        text = [
            "import os",
            "import numpy as np",
            "from tqdm import tqdm",
            "from inference import Classifier",
            "def main():",
            "    data_dir = '.'",
            "    output_dir = './output'",
            "    if 'output' not in os.listdir():",
            "        os.mkdir('./output')",
            "    files = {",
            "        'h':os.path.join(data_dir, 'h.fits'),",
            "        'j':os.path.join(data_dir, 'j.fits'),",
            "        'v':os.path.join(data_dir, 'v.fits'),",
            "        'z':os.path.join(data_dir, 'z.fits')",
            "    }",
            "    Classifier.classify_files(h=files['h'],",
            "                              j=files['j'],",
            "                              v=files['v'],",
            "                              z=files['z'],",
            "                              batch_size=2000,",
            "                              out_type='rank_vote',",
            "                              out_dir=output_dir)",
            "if __name__=='__main__':",
            "    main()",
        ]

        with open(os.path.join(path, "main.py"), "w") as f:
            f.write("\n".join(text))

    @staticmethod
    def _stitch_parallel_classifications(out_dir: str) -> None:
        """Stitch the seperate outputs made from the parallel classifications.

        Args:
            out_dir (str): the location that contains the parallel classified
                           subdirs

        Returns:
            None
        """

        for f in helpers.LabelHelper.MORPHOLOGIES:
            to_be_stitched = []
            for output in sorted(os.listdir(out_dir)):
                if os.path.isdir(output):
                    fname = os.path.join(output, "output/{}.fits".format(f))
                    to_be_stitched.append(fits.getdata(fname)[-1, :, :])

            size = to_be_stitched[0].shape
            new_y = sum(t.shape[0] for t in to_be_stitched) - (
                40 * (len(to_be_stitched) - 1)
            )
            new_x = size[1]
            combined = np.zeros(shape=[new_y, new_x], dtype=np.float32)
            start_y = 0
            for t in to_be_stitched:
                combined[start_y : start_y + t.shape[0], :] += t
                start_y = start_y + t.shape[0] - 40

            fits.PrimaryHDU(data=combined).writeto(
                os.path.join(out_dir, f"{f}.fits"), overwrite=True
            )

    @staticmethod
    def _build_parallel_classification_structure(
        arrs: List[np.ndarray], gpus: List[int], out_dir: str
    ) -> None:
        """Sets up the subdirs and files to run the parallel classification.

        Args:
            arrs (List[np.ndarray]): List of arrays to split up in the order HJVZ
            gpus (List[int]): A list of the CUDA gpu ID's to use for a
                              parallel classification.
            out_dir (str): the location to place the subdirs in

        Returns:
            None
        """

        shape = arrs[0].shape
        num_gpus = len(gpus)
        split_slices = Classifier._get_split_slice_generator(
            shape, num_gpus, Classifier._get_split_length(shape, num_gpus)
        )

        for gpu, split_slice in tqdm(zip(sorted(gpus), split_slices)):
            sub_output_dir = os.path.join(out_dir, str(gpu))
            os.mkdir(sub_output_dir)

            for name, data in zip(["h", "j", "v", "z"], arrs):
                tmp_location = os.path.join(sub_output_dir, "{}.fits".format(name))
                fits.PrimaryHDU(data=data[split_slice, :]).writeto(tmp_location)

            Classifier._make_runnable_file(sub_output_dir)

    @staticmethod
    def _run_parallel_jobs(
        gpus: List[int], out_dir: str, parallel_check_interval: int
    ) -> None:
        """Starts and tracks parallel job runs.

        Warning: This will not finish running until all subprocesses are complete

        Args:
            gpus (List[int]): A list of the CUDA gpu ID's to use for a
                              parallel classification.
            out_dir (str): the location with the partitioned data
            parallel_check_interval (int): If gpus are given, then this is the number
                                           of minutes to wait between polling each
                                           subprocess for completetion

        Returns:
            None
        """

        processes = {}

        for gpu in gpus:
            cmd_string = f"CUDA_VISIBLE_DEVICES={gpu} python main.py"
            sub_dir = os.path.join(out_dir, str(gpu))
            processes[gpu] = Popen(cmd_string, shell=True, cwd=sub_dir)

        is_running = np.ones([len(gpus)], dtype=np.bool)

        while is_running.any():
            for i, g in enumerate(sorted(gpus)):
                if is_running[i] and processes[g].poll():
                    is_running[i] = False

            time.sleep(parallel_check_interval * 60)

    @staticmethod
    def colorize_rank_vote_output(
        data: dict, out_dir: str = None, hide_unclassified: bool = True
    ) -> np.ndarray:
        """Makes a color images from the rank_vote classification output.

        The colorization scheme is defined in HSV and is as follows:

        * Spheroid = Red
        * Disk = Blue
        * Irregular = Green
        * Point Source = Yellow

        The hue is set to be the color associated with the highest ranked class
        for a given pixel. The saturation is set to be difference between the
        highest ranked class and the second highest ranked class for a given
        pixel. For example if the top two classes have nearly equal values given
        by the classifier, then the saturation will be low and the pixel will
        appear more white. If the the top two classes have very different
        values, then the saturation will be high and the pixel's color will be
        vibrant and not white. The value for a pixel is set to be 1-bkg, where
        bkg is value given to the background class. If the background class has
        a high value, then the pixel will appear more black. If the background
        value is low, then the pixel will take on the color given by the hue and
        saturation values.

        Args:
            data (dict): A dictionary containing the output from Morpheus.
            out_dir (str): a path to save the image in.
            hide_unclassified (bool): If true black out the edges of the image
                                      that are unclassified. If false, show the
                                      borders as white.

        Returns:
            A [width, height, 3] array representing the RGB image.
        """
        red = 0.0  # spheroid
        blue = 0.7  # disk
        yellow = 0.18  # point source
        green = 0.3  # irregular

        shape = data["n"].shape

        colors = np.array([red, blue, green, yellow])
        morphs = np.dstack([data[i] for i in helpers.LabelHelper.MORPHOLOGIES[:-1]])
        ordered = np.argsort(-morphs, axis=-1)

        hues = np.zeros(shape)
        sats = np.zeros(shape)
        vals = 1 - data["background"]

        # the classifier doesn't return values for this area so black it out
        if hide_unclassified:
            vals[0:5, :] = 0
            vals[-5:, :] = 0
            vals[:, 0:5] = 0
            vals[:, -5:] = 0

        for i in tqdm(range(shape[0])):
            for j in range(shape[1]):
                hues[i, j] = colors[ordered[i, j, 0]]
                sats[i, j] = (
                    morphs[i, j, ordered[i, j, 0]] - morphs[i, j, ordered[i, j, 1]]
                )

        hsv = np.dstack([hues, sats, vals])
        rgb = hsv_to_rgb(hsv)

        if out_dir:
            png = (rgb * 255).astype(np.uint8)
            imageio.imwrite(os.path.join(out_dir, "colorized.png"), png)

        return rgb

    @staticmethod
    def make_segmap(
        data: dict,
        h: np.ndarray,
        out_dir: str = None,
        min_distance: int = 20,
        mask: np.ndarray = None,
        deblend: bool = True,
    ) -> np.ndarray:
        """Generate a segmentation map from the classification output.

            Segmentation maps are generated using the background classification
            in ``data``. The segmenatation maps are generated using the
            `watershed algorithm <https://en.wikipedia.org/wiki/Watershed_(image_processing)>`_
            in conjunction with a couple heuristics for simple deblending. The
            steps to generate the segmentation map are as follows:

            **Inputs**: Background pixel classifications *b* and H band flux *h*

            **Outpus**: A segmentation map *sm*

            **Algorithm**:

            1. Initial Segmentation

            :math:`psf_r` = The radius of the PSF for the instrument used in H band

            *m* = a zero matrix, same size as *b*

            FOR :math:`m_{ij}` in *m*

                IF :math:`b_{ij}==1` THEN :math:`m_{ij}=1`

                ELSE IF :math:`b_{ij}==0` THEN :math:`m_{ij}=2`

            *s* = `sobel <https://en.wikipedia.org/wiki/Sobel_operator>`_ (*b*)

            *sm* = apply waterhshed on *s* with markers *m*

            2. Simple Deblending

            FOR EACH contiguous set of source pixels *p* in *sm* and corresponding flux set *f* in *h*

                Assign a unique id to *p*

                *mx* = :math:`\\max(f)`

                *pm* = pixels in *p* that are at least :math:`\\frac{mx}{10}` AND at least a :math:`psf_r` apart.

                IF there are more than 2 pixels *pm*

                    *pwm* = apply watershed on :math:`-1 * f` with markers *pm*

                    Assign unique ids to the contiguous sets of pixels in *pwm*

            RETURN *sm*

        Args:
            data (dict): A dictionary containing the output from morpheus.
            h (np.ndarray): The H band image that was classified.
            out_dir (str): A path to save the segmap in.
            min_distance (int): The minimum distance for deblending
            mask (np.ndarry): A boolean mask indicating which pixels
            deblend (bool): If ``True``, perform delblending as described in 2.
                            in the algorithm description. If ``False`` return
                            segmap without deblending.

        Returns:
            A np.ndarray segmentation map
        """
        bkg = data["background"]
        markers = np.zeros_like(h, dtype=np.uint8)

        print("Building Markers...")
        if mask is None:
            mask = data["n"] > 0

        is_bkg = np.logical_and(bkg == 1, mask)
        is_src = np.logical_and(bkg == 0, mask)

        markers[is_bkg] = 1
        markers[is_src] = 2

        sobel_img = sobel(bkg)

        print("Watershedding...")
        segmented = watershed(sobel_img, markers, mask=mask) - 1
        segmented[np.logical_not(mask)] = 0

        labeled, _ = ndi.label(segmented)

        labeled[np.logical_not(mask)] = -1

        if deblend:
            labeled = Classifier._deblend(labeled, h, min_distance)

        if out_dir:
            fits.PrimaryHDU(data=labeled).writeto(os.path.join(out_dir, "segmap.fits"))

        return labeled

    @staticmethod
    def _deblend(segmap: np.ndarray, flux: np.ndarray, min_distance: int) -> np.ndarray:
        """Deblends a segmentation map according to the description in make_segmap.

        Args:
            segmap (np.ndarray): The segmentation map image to deblend
            flux (np.ndarray): The corresponding flux image in H band
            min_distance (int): The radius of the PSF for the instrument used on H band

        Returns:
            A np.ndarray representing the deblended segmap
        """

        max_id = segmap.max()

        for region in tqdm(regionprops(segmap, flux), desc="Deblending"):

            # greater than 1 indicates that the region is not background
            if region.label > 0:
                flx = region.intensity_image
                seg = region.filled_image
                flux_map = flx * seg

                maxes = peak_local_max(
                    flux_map, min_distance=min_distance, num_peaks=20
                )

                # more than 1 source found, deblend
                if maxes.shape[0] > 1:
                    start_y, start_x, end_y, end_x = region.bbox
                    markers = np.zeros_like(seg, dtype=np.int)

                    for y, x in maxes:
                        max_id += 1
                        markers[y, x] = max_id

                    deblended = watershed(-flux_map, markers, mask=seg)

                    local_segmap = segmap[start_y:end_y, start_x:end_x].copy()
                    local_segmap = np.where(seg, deblended, local_segmap)
                    segmap[start_y:end_y, start_x:end_x] = local_segmap

        return segmap

    @staticmethod
    def aggregation_scheme_flux_weighted(
        data: dict, flux: np.ndarray, segmap: np.ndarray
    ) -> List[float]:
        """Aggreates pixel level morphological classifcations to the source level.

        Uses a flux weighted mean of the pixel level morphologies to calculate
        the aggregate source level morphology.

        Args:
            data (dict): A dictionary containing the output from morpheus.
            flux (np.ndarray): The corresponding flux image in H band
            segmap (int): The binary map indicating pixels that belong to the
                          source

        Returns:
            The morphological classification as a list of floats in the
            following order: ['spheroid', 'disk', 'irregular', 'point source']
        """
        classifications = np.zeros([4])

        morphs = ["spheroid", "disk", "irregular", "point_source"]

        morphs = [data[m] for m in morphs]

        for i, m in enumerate(morphs):
            classifications[i] = np.mean(m[segmap] * flux[segmap])

        return (classifications / classifications.sum()).tolist()

    @staticmethod
    def make_catalog(
        data: dict,
        flux: np.ndarray,
        segmap: np.ndarray,
        aggregation_scheme: Callable = None,
    ) -> List[Dict]:
        """Creates a catalog of sources and their morphologies.

        Args:
            data (dict): A dictionary containing the output from morpheus.
            flux (np.ndarray): The corresponding flux image in H band
            segmap (int): A labelled segmap where every pixel with a
                          value > 0 is associated with a source.
            aggregation_scheme (func): Function that takes three arguments `data`,
                                       `flux`, and `segmap`, same as to this
                                       function, then returns a numpy array
                                       containing the morphological classification
                                       in the following order-shperoid, disk,
                                       irregular, and point source/compact. If
                                       None, then the flux weighting scheme
                                       in

        Returns:
            A JSON-compatible list of dictionary objects with the following keys:
            {
                'id': the id from the segmap
                'location': a (y,x) location -- the max pixel within the segmap
                'morphology': a dictionary containing the morphology values.
            }
        """

        if aggregation_scheme is None:
            aggregation_scheme = Classifier.aggregation_scheme_flux_weighted

        catalog = []

        for region in regionprops(segmap, flux):
            _id = region.label

            if _id < 1:
                continue

            img = region.intensity_image
            seg = region.filled_image

            start_y, start_x, end_y, end_x = region.bbox
            dat = {}
            for k in data:
                dat[k] = data[k][start_y:end_y, start_x:end_x].copy()

            classification = aggregation_scheme(dat, img, seg)

            masked_flux = img * seg

            # https://stackoverflow.com/a/3584260
            local_y, local_x = np.unravel_index(masked_flux.argmax(), masked_flux.shape)
            global_y, global_x = int(start_y + local_y), int(start_x + local_x)

            catalog.append(
                {
                    "id": _id,
                    "location": [global_y, global_x],
                    "morphology": classification,
                }
            )

        return catalog



# MIT License
# Copyright 2018 Ryan Hausen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ==============================================================================
"""A base class for building neural network models in Tensorflow."""

from types import FunctionType

import tensorflow as tf

from morpheus.core.helpers import OptionalFunc


class Model:
    """ Base class for models.

    Attributes:
        dataset (tf.data.Dataset): Dataset Object for training
        is_training (bool): indicates whether or not the model is training
        data_format (str): 'channels_first' or 'channels_last'

    Required methods to override:
        model_fn: the graph function

    Optional methods to override:
        train_metrics: to add metrics during training
        test_metrics: to add metrics during testing, can be same as train_metrics
        optimizer: updates params based on a loss tensor
        loss_func: defines a loss value given and x and y tensor
        inference: default applies softmax to tensor from model_fn
    """

    train_metrics = OptionalFunc("No training metrics set")
    test_metrics = OptionalFunc("No test metrics set")
    optimizer = OptionalFunc("No optimizer set")
    loss_func = OptionalFunc("No loss function set")
    inference = OptionalFunc("Nor inference fucntion set")

    def __init__(self, dataset: tf.data.Dataset, data_format: str = "channels_last"):
        """Inits Model with dataset, and data_format"""
        self.dataset = dataset
        self.data_format = data_format
        self._graph = None

    def model_fn(self, inputs: tf.Tensor, is_training: bool) -> FunctionType:
        """Function that defines model. Needs to be Overidden!

        Args:
            inputs (tf.Tensor): the input tensor
            is_training (bool): boolean to indicate if in training phase

        Returns:
            Should return a function that takes two inputs tf.Tensor and bool

        Raises:
            NotImplementedError if not overridden
        """
        raise NotImplementedError()

    def build_graph(self, inputs: tf.Tensor, is_training: bool) -> tf.Tensor:
        """Base function that returns model_fn evaluated on x. Don't Override!

        Args:
            inputs (tf.Tensor): The tensor to be processed, ie a placeholder
            is_training (bool): whether or not the model is training useful
                                for things like batch normalization or
                                dropout

        Returns:
            returns the tensor that represents the result of model_fn evaluated
            on the input tensor

        Raises
            NotImplementedError if Model.model_fn() is not overwritten
        """
        if self._graph:
            return self._graph(inputs, is_training)

        self._graph = self.model_fn
        return self._graph(inputs, is_training)

    def train(self) -> (tf.Tensor, tf.Tensor):
        """Builds the training routine tensors. Don't Override!

        Returns:
            (optimize, metrics): the result of self.optimizer and
                                 self.train_metrics respectively

        Raises
            NotImplementedError if Model.model_fn() is not overwritten
        """
        data, labels = self.dataset.train
        logits = self.build_graph(data, True)
        optimize = self.optimizer(self.loss_func(logits, labels))
        metrics = self.train_metrics(logits, labels)

        return optimize, metrics

    def test(self) -> (tf.Tensor, tf.Tensor):
        """Builds the testing routing tensors. Don't Override!

        Returns:
            (logits, metrics): the result of the self.build_graph and
                               self.test_metrics respectively

        Raises
            NotImplementedError if Model.model_fn() is not overwritten
        """
        inputs, labels = self.dataset.test
        logits = self.build_graph(inputs, False)
        metrics = self.test_metrics(logits, labels)

        return logits, metrics



# MIT License
# Copyright 2018 Ryan Hausen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ==============================================================================
"""Tests morpheus.classifier module."""
import os

import pytest
import numpy as np

from morpheus.classifier import Classifier
from morpheus.data import example

import morpheus.tests.data_helper as dh


@pytest.mark.unit
class TestClassifier:
    """Tests morpheus.classifier.Classifier"""

    @staticmethod
    def test_variables_not_none():
        """Tests _variables_not_none method."""

        names = ["a", "b", "c"]
        values = [1, 2, 3]

        Classifier._variables_not_none(names, values)

    @staticmethod
    def test_variables_not_none_throws():
        """Tests _variables_not_none, throws ValueError."""

        names = ["a", "b", "c"]
        values = [1, 2, None]

        with pytest.raises(ValueError):
            Classifier._variables_not_none(names, values)

    @staticmethod
    def test_arrays_same_size():
        """Tests _arrays_same_size method."""

        shape = (10, 10)
        arrs = [np.zeros(shape) for _ in range(3)]

        Classifier._arrays_same_size(arrs)

    @staticmethod
    def test_arrays_same_size_throws():
        """Tests _arrays_same_size, throws ValueError."""

        shape = (10, 10)
        arrs = [np.zeros(shape) for _ in range(3)]
        arrs.append(np.zeros((20, 20)))

        with pytest.raises(ValueError):
            Classifier._arrays_same_size(arrs)

    @staticmethod
    def test_standardize_img():
        """Test _standardize_img method."""

        img = np.random.normal(loc=1.0, scale=3.0, size=(10, 10, 10))

        img = Classifier._standardize_img(img)

        np.testing.assert_allclose(np.mean(img), 0, atol=1e-07)
        np.testing.assert_allclose(np.var(img), 1, atol=1e-07)

    @staticmethod
    def test_make_runnable_file():
        """Test _make_runnable_file."""
        local = os.path.dirname(os.path.abspath(__file__))

        Classifier._make_runnable_file(local)

        assert os.path.exists(os.path.join(local, "main.py"))

        os.remove(os.path.join(local, "main.py"))

    @staticmethod
    def test_variables_not_none_raises():
        """Test _variables_not_none."""

        with pytest.raises(ValueError):
            Classifier._variables_not_none(["good", "bad"], [1, None])

    @staticmethod
    def test_colorize_rank_vote_output():
        """Test colorize_rank_vote_output."""

        data = dh.get_expected_morpheus_output()
        expected_color = dh.get_expected_colorized_pngs()["no_hidden"]

        actual_color = Classifier.colorize_rank_vote_output(
            data, hide_unclassified=False
        )

        actual_color = (actual_color * 255).astype(np.uint8)

        np.testing.assert_array_almost_equal(expected_color, actual_color)

    @staticmethod
    def test_colorize_rank_vote_output_hidden():
        """Test colorize_rank_vote_output with hiding."""

        data = dh.get_expected_morpheus_output()
        expected_color = dh.get_expected_colorized_pngs()["hidden"]

        actual_color = Classifier.colorize_rank_vote_output(
            data, hide_unclassified=True
        )

        actual_color = (actual_color * 255).astype(np.uint8)

        np.testing.assert_array_almost_equal(expected_color, actual_color)

    @staticmethod
    def test_make_segmap():
        """Test make_segmap method."""

        data = dh.get_expected_morpheus_output()
        h, _, _, _ = example.get_sample()
        mask = np.zeros_like(h, dtype=np.int)
        mask[5:-5, 5:-5] = 1

        expected_segmap = dh.get_expected_segmap()["segmap"]

        actual_segmap = Classifier.make_segmap(data, h, mask=mask)

        np.testing.assert_array_equal(expected_segmap, actual_segmap)



# MIT License
# Copyright 2018 Ryan Hausen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ==============================================================================
"""Used to fetch data files"""

import os
from typing import List
from astropy.io import fits
import numpy as np


def get_sample(out_dir: str = None) -> List[np.ndarray]:
    """Retrieves a sample of CANDELS data as an example for classification.

    Sample image is taken from CANDELS 1.0 release. The object in the center
    is GDS_deep2_3622 from Karaltepe et. al (2015).

    Data can be manually downlaoded from:

    https://drive.google.com/uc?export=download&id=1fFGUVOVMOGLG4ptgAZv0T9Woimr653kn

    Args:
        out_dir (str): if not None the location to save the data at.

    Returns:
        A list of numpy arrays containing the H, J, V, and Z bands in that order

    """
    # Got direct link format from: https://www.labnol.org/internet/direct-links-for-google-drive/28356/
    url = "https://drive.google.com/uc?export=download&id=1fFGUVOVMOGLG4ptgAZv0T9Woimr653kn"

    raw_data = fits.getdata(url)
    bands = [raw_data[i] for i in range(4)]

    if out_dir:
        for name, data in zip(["h", "j", "v", "z"], bands):
            out_path = os.path.join(out_dir, f"{name}.fits")
            fits.PrimaryHDU(data=data).writeto(out_path)

    return bands



# MIT License
# Copyright 2018 Ryan Hausen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ==============================================================================
"""Tests the helper functions"""

import os
import pytest
import numpy as np
import tensorflow as tf
from astropy.io import fits

from morpheus.core.helpers import FitsHelper
from morpheus.core.helpers import TFLogger
from morpheus.core.helpers import OptionalFunc
from morpheus.core.helpers import LabelHelper


@pytest.mark.unit
class TestTFLogger:
    """This tests the TFLogger class' funtions.

    TODO: Figure out to properly test this.
    """

    @staticmethod
    def test_info():
        """Tests TFlogger.info"""
        TFLogger.info("Test Message")

    @staticmethod
    def test_debug():
        """Tests TFlogger.debug"""
        TFLogger.debug("Test Message")

    @staticmethod
    def test_warn():
        """Tests TFlogger.warn"""
        TFLogger.warn("Test Message")

    @staticmethod
    def test_error():
        """Tests TFlogger.error"""
        TFLogger.error("Test Message")

    @staticmethod
    def test_tensor_shape():
        """Tests TFlogger.tensor_shape"""
        t = tf.zeros([3, 3], dtype=tf.int32, name="test")

        TFLogger.tensor_shape(t)


@pytest.mark.unit
class TestOptionalFunc:
    """This tests OptionalFunc class."""

    @staticmethod
    def test_placeholder():
        """Tests the dummy function in OptionalFunc"""
        expected_args = [1, 2, 3]

        opt = OptionalFunc("Test")

        assert expected_args == opt.placeholder(1, 2, 3)

    @staticmethod
    def test_default_is_set():
        """Tests that the default flag is set on init for OptionalFunc"""

        of = OptionalFunc("Test")
        of.__get__(None)
        assert of._is_default

    @staticmethod
    def test_default_is_set_off():
        """Tests that the default flag is set on init for OptionalFunc"""

        of = OptionalFunc("Test")
        of.__set__(None, lambda x: x)

        assert not of._is_default


@pytest.mark.unit
class TestFitsHelper:
    """This tests the FITS helper class."""

    @staticmethod
    def test_create_file_rasies():
        """Tests that a ValueError is raised from improper dtype param"""

        with pytest.raises(ValueError):
            FitsHelper.create_file("dummy.fits", [0], str)

    @staticmethod
    def test_create_unit8_file():
        """Tests 2d uint8 file creation"""

        local = os.path.dirname(os.path.abspath(__file__))
        dummy_name = os.path.join(local, "dummy_unit8.fits")
        dummy_shape = (200, 200)
        dummy_dtype = np.uint8

        FitsHelper.create_file(dummy_name, dummy_shape, dummy_dtype)

        retrieved = fits.getdata(dummy_name)

        os.remove(dummy_name)

        assert dummy_shape == retrieved.shape
        assert np.issubdtype(dummy_dtype, retrieved.dtype)

    @staticmethod
    def test_create_int16_file():
        """Tests 2d int16 file creation"""

        local = os.path.dirname(os.path.abspath(__file__))
        dummy_name = os.path.join(local, "dummy_int16.fits")
        dummy_shape = (200, 200)
        dummy_dtype = np.int16

        FitsHelper.create_file(dummy_name, dummy_shape, dummy_dtype)

        retrieved = fits.getdata(dummy_name)

        os.remove(dummy_name)

        assert dummy_shape == retrieved.shape
        assert np.issubdtype(dummy_dtype, retrieved.dtype)

    @staticmethod
    def test_create_int32_file():
        """Tests 2d int32 file creation"""

        local = os.path.dirname(os.path.abspath(__file__))
        dummy_name = os.path.join(local, "dummy_int32.fits")
        dummy_shape = (200, 200)
        dummy_dtype = np.int32

        FitsHelper.create_file(dummy_name, dummy_shape, dummy_dtype)

        retrieved = fits.getdata(dummy_name)

        os.remove(dummy_name)

        assert dummy_shape == retrieved.shape
        assert np.issubdtype(dummy_dtype, retrieved.dtype)

    @staticmethod
    def test_create_float32_file():
        """Tests 2d float32 file creation"""

        local = os.path.dirname(os.path.abspath(__file__))
        dummy_name = os.path.join(local, "dummy_float32.fits")
        dummy_shape = (200, 200)
        dummy_dtype = np.float32

        FitsHelper.create_file(dummy_name, dummy_shape, dummy_dtype)

        retrieved = fits.getdata(dummy_name)

        os.remove(dummy_name)

        assert dummy_shape == retrieved.shape
        assert np.issubdtype(dummy_dtype, retrieved.dtype)

    @staticmethod
    def test_create_float64_file():
        """Tests 2d float64 file creation"""

        local = os.path.dirname(os.path.abspath(__file__))
        dummy_name = os.path.join(local, "dummy_float64.fits")
        dummy_shape = (200, 200)
        dummy_dtype = np.float64

        FitsHelper.create_file(dummy_name, dummy_shape, dummy_dtype)

        retrieved = fits.getdata(dummy_name)

        os.remove(dummy_name)

        assert dummy_shape == retrieved.shape
        assert np.issubdtype(dummy_dtype, retrieved.dtype)

    @staticmethod
    def test_create_file_3d():
        """Tests a 3d file creation"""

        local = os.path.dirname(os.path.abspath(__file__))
        dummy_name = os.path.join(local, "dummy_unit8_3d.fits")
        dummy_shape = (200, 200, 5)
        dummy_dtype = np.uint8

        FitsHelper.create_file(dummy_name, dummy_shape, dummy_dtype)

        retrieved = fits.getdata(dummy_name)

        os.remove(dummy_name)

        # the channels when stored as fits get moved to the front.
        dummy_shape = tuple(reversed(dummy_shape))
        assert dummy_shape == retrieved.shape
        assert np.issubdtype(dummy_dtype, retrieved.dtype)

    @staticmethod
    def test_get_files():
        """Tests the get_files() method"""

        local = os.path.dirname(os.path.abspath(__file__))

        dummy_file1 = os.path.join(local, "dummy1.fits")
        dummy_file2 = os.path.join(local, "dummy2.fits")

        arr1 = np.zeros([100, 100], dtype=np.float32)
        arr2 = np.ones([100, 100], dtype=np.float32)

        for f, a in zip([dummy_file1, dummy_file2], [arr1, arr2]):
            fits.PrimaryHDU(data=a).writeto(f)

        _, arrs = FitsHelper.get_files([dummy_file1, dummy_file2])

        os.remove(dummy_file1)
        os.remove(dummy_file2)

        assert np.array_equal(arr1, arrs[0])
        assert np.array_equal(arr2, arrs[1])

    @staticmethod
    def test_create_mean_var_files():
        """Tests create_mean_var_files method."""

        morphs = ["spheroid", "disk", "irregular", "point_source", "background"]

        local = os.path.dirname(os.path.abspath(__file__))

        shape = (200, 200)
        expected_file_names = []

        for m in morphs:
            for t in ["mean", "var"]:
                expected_file_names.append(os.path.join(local, f"{m}_{t}.fits"))

        FitsHelper.create_mean_var_files(shape, local)

        for f in expected_file_names:
            arr = fits.getdata(f)
            assert arr.shape == shape
            assert np.issubdtype(np.float32, arr.dtype)
            os.remove(f)

    @staticmethod
    def test_create_rank_vote_files():
        """Tests create_rank_vote_files method."""

        morphs = ["spheroid", "disk", "irregular", "point_source", "background"]

        local = os.path.dirname(os.path.abspath(__file__))

        shape = (200, 200)
        expected_file_names = []

        for m in morphs:
            expected_file_names.append(os.path.join(local, f"{m}.fits"))

        FitsHelper.create_rank_vote_files(shape, local)

        for f in expected_file_names:
            arr = fits.getdata(f)
            assert arr.shape == shape
            assert np.issubdtype(np.float32, arr.dtype)
            os.remove(f)

    @staticmethod
    def test_create_n_file():
        """Test create_n_file method."""

        local = os.path.dirname(os.path.abspath(__file__))

        shape = (200, 200)
        expected_file_name = os.path.join(local, "n.fits")

        FitsHelper.create_n_file(shape, local)

        arr = fits.getdata(expected_file_name)
        assert arr.shape == shape
        assert np.issubdtype(np.int16, arr.dtype)

        os.remove(expected_file_name)


@pytest.mark.unit
class TestLabelHelper:
    """This tests the LabelHelper class.

    TODO: Integration test for windowed_index_generator & final_map
    """

    @staticmethod
    def test_index_generator():
        """Test the index_generator method."""
        shape = (100, 100)

        arr = np.zeros(shape, dtype=np.bool)

        index_gen = LabelHelper.index_generator(*shape)

        for i, j in index_gen:
            arr[i, j] = True

        assert arr.all()

    @staticmethod
    def test_windowed_index_generator():
        """Test the windowed_index_generator function."""

        shape = (100, 100)

        arr = np.zeros(shape, dtype=np.bool)

        index_gen = LabelHelper.windowed_index_generator(*shape)

        for i, j in index_gen:
            arr[i, j] = True

        window_y, window_x = LabelHelper.UPDATE_MASK_N.shape
        expected_array = np.ones((shape[0] - window_y + 1, shape[1] - window_x + 1))

        expected_array = np.pad(
            expected_array, ((0, window_y - 1), (0, window_x - 1)), "constant"
        ).astype(np.bool)

        np.testing.assert_array_equal(arr, expected_array)

    @staticmethod
    def test_get_final_map_complete():
        """Test the get_final_map method on complete array."""

        shape = (100, 100)
        arr = np.zeros(shape, dtype=np.bool)

        for i in range(shape[0] - LabelHelper.UPDATE_MASK_N.shape[0] + 1):
            for j in range(shape[1] - LabelHelper.UPDATE_MASK_N.shape[1] + 1):
                final_map = LabelHelper.get_final_map(shape, i, j)
                for (y, x) in final_map:
                    arr[i + y, j + x] = True

        expected_array = np.zeros(shape, dtype=np.bool)
        expected_array[5:-5, 5:-5] = True

        all_same = np.equal(expected_array, arr)

        assert all_same.all()

    @staticmethod
    def test_iterative_mean():
        """Test the iterative_mean method."""

        shape = (10, 10)
        terms = [np.ones(shape) * i for i in range(9)]
        mean = np.zeros(shape)

        for i in range(9):
            ns = np.ones(shape) * (i + 1)
            mean = LabelHelper.iterative_mean(ns, mean, terms[i], np.ones_like(mean))

        expected_mean = np.mean(terms, axis=0)
        all_same = np.equal(expected_mean, mean)

        assert all_same.all()

    @staticmethod
    def test_iterative_variance_no_final():
        """Test the iterative variance before finalization."""

        shape = (10, 10)
        terms = [np.ones(shape) * i for i in range(9)]
        s_n = np.zeros(shape)
        update_mask = np.ones((shape))

        for i in range(9):
            curr_mean = np.mean(terms[: i + 1], axis=0)
            if i > 0:
                prev_mean = np.mean(terms[:i], axis=0)
            else:
                prev_mean = curr_mean.copy()

            s_n = LabelHelper.iterative_variance(
                s_n, terms[i], prev_mean, curr_mean, update_mask
            )

        n = np.ones(shape) * 9
        expected_sn = np.var(terms, axis=0) * n

        all_same = np.equal(expected_sn, s_n)

        assert all_same.all()

    @staticmethod
    def test_iterative_variance_with_final():
        """Test the iterative variance with finalization."""

        shape = (10, 10)
        terms = [np.ones(shape) * i for i in range(9)]

        expected_var = np.var(terms, axis=0)

        n = np.ones(shape) * 9
        sn = expected_var * n

        final_map = []

        for i in range(shape[0]):
            for j in range(shape[1]):
                final_map.append((i, j))

        var = LabelHelper.finalize_variance(n, sn, final_map)

        all_same = np.equal(expected_var, var)

        assert all_same.all()

    @staticmethod
    def test_iterative_rank_vote():
        """Test the iterative_rank_vote method."""

        shape = (10, 10)
        prev_votes = np.zeros(shape)

        ranks = np.zeros(shape, dtype=int)
        ranks[:, 0] = 4

        update_mask = np.ones(shape)
        votes = LabelHelper.iterative_rank_vote(ranks, prev_votes, update_mask)

        expected_votes = prev_votes.copy()
        expected_votes[:, 0] = 1

        all_same = np.equal(expected_votes, votes)

        assert all_same.all()

    @staticmethod
    def test_make_mean_var_arrays():
        """Test the make_mean_var_arrays method."""

        expected_shape = (100, 100)
        expected_dtype = np.float32

        expected_keys = []
        for morph in LabelHelper.MORPHOLOGIES:
            for t in ["mean", "var"]:
                expected_keys.append(f"{morph}_{t}")

        outs = LabelHelper.make_mean_var_arrays(expected_shape)

        for k in expected_keys:
            assert expected_shape == outs[k].shape
            assert np.issubdtype(expected_dtype, outs[k].dtype)

    @staticmethod
    def test_make_rank_vote_arrays():
        """Test the make_rank_vote_arrays method."""

        expected_shape = (100, 100)
        expected_dtype = np.float32
        expected_keys = LabelHelper.MORPHOLOGIES

        outs = LabelHelper.make_rank_vote_arrays(expected_shape)

        for k in expected_keys:
            assert expected_shape == outs[k].shape
            assert np.issubdtype(expected_dtype, outs[k].dtype)

    @staticmethod
    def test_make_n_array():
        """Test the make_n_array method."""

        expected_shape = (100, 100)
        expected_dtype = np.int16
        expected_keys = "n"

        outs = LabelHelper.make_n_array(expected_shape)

        for k in expected_keys:
            assert expected_shape == outs[k].shape
            assert np.issubdtype(expected_dtype, outs[k].dtype)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Adapted from: https://github.com/kennethreitz/setup.py

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = "morpheus-astro"
DESCRIPTION = "A Library For Generating Morphological Semantic Segmentation Maps of Astronomical Images"
URL = "https://github.com/morpheus-project/morpheus"
EMAIL = "rhausen@ucsc.edu"
AUTHOR = "Ryan Hausen & Brant Robertson"
REQUIRES_PYTHON = ">=3.6"
VERSION = "0.3.1"

# What packages are required for this module to be executed?
REQUIRED = [
    "numpy<1.16",  # remove retriction once skimage >= 14.2
    "colorama",
    "astropy",
    "tqdm",
    "matplotlib",
    "imageio",
    "scikit-image",
    "scipy",
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(VERSION))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=("morpheus.tests",)),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: PyPy",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering",
    ],
    # $ setup.py publish support.
    # cmdclass={"upload": UploadCommand},
)



# MIT License
# Copyright 2018 Ryan Hausen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ==============================================================================
"""Implements variations of the U-Net architecture."""

import tensorflow as tf
import morpheus.core.base_model


LAYERS = tf.layers
VarInit = tf.variance_scaling_initializer


class Model(morpheus.core.base_model.Model):
    """Based on U-Net (https://arxiv.org/abs/1505.04597).

    Args:
        hparams (tf.contrib.training.HParams): Hyperparamters to use
        dataset (tf.data.Dataset): dataset to use for training
        data_format (str): channels_first or channels_last

    Required HParams:
        down_filters (list): number of filters for each down conv section
        num_down_convs (int): number of conv ops per down conv section
        up_filters (list): number of filters for each up conv section
        num_up_convs (int): number of conv ops per up conv section
        batch_norm (bool): use batch normalization
        dropout (bool): use dropout

    Optional HParams:
        dropout_rate (float): the percentage of neurons to drop [0.0, 1.0]
    """

    def __init__(
        self,
        hparams: tf.contrib.training.HParams,
        dataset: tf.data.Dataset,
        data_format="channels_last",
    ):
        """Inits Model with hparams, dataset, and data_format"""

        super().__init__(dataset, data_format)
        self.hparams = hparams

    def model_fn(self, inputs: tf.Tensor, is_training: bool) -> tf.Tensor:
        """Defines U-Net graph using HParams.

        Args:
            inputs (tf.Tensor): The input tensor to the graph
            is_training (bool): indicates if the model is in the training phase

        Returns:
            tf.Tensor: the output tensor from the graph

        TODO: add input shape check for incomaptible tensor shapes
        """

        outputs = []

        for idx, num_filters in enumerate(self.hparams.down_filters):
            with tf.variable_scope("downconv-{}".format(idx), reuse=tf.AUTO_REUSE):
                for c_idx in range(self.hparams.num_down_convs):
                    with tf.variable_scope(
                        "conv-{}".format(c_idx), reuse=tf.AUTO_REUSE
                    ):
                        inputs = self.block_op(inputs, num_filters, is_training)

                outputs.append(inputs)
                inputs = self.down_sample(inputs)

        with tf.variable_scope("intermediate-conv", reuse=tf.AUTO_REUSE):
            inputs = self.block_op(
                inputs, self.hparams.num_intermediate_filters, is_training
            )

        concat_axis = 3 if self.data_format == "channels_last" else 1
        for idx, num_filters in enumerate(self.hparams.up_filters):
            with tf.variable_scope("upconv-{}".format(idx), reuse=tf.AUTO_REUSE):
                inputs = self.up_sample(inputs)
                inputs = tf.concat(
                    [inputs, outputs[-(idx + 1)]], concat_axis, name="concat_op"
                )
                for c_idx in range(self.hparams.num_up_convs):
                    with tf.variable_scope(
                        "conv-{}".format(c_idx), reuse=tf.AUTO_REUSE
                    ):
                        inputs = self.block_op(inputs, num_filters, is_training)

        with tf.variable_scope("final_conv", reuse=tf.AUTO_REUSE):
            inputs = self.conv(
                inputs, self.dataset.num_labels, activation=None, kernel_size=3
            )

        return inputs

    def block_op(
        self, inputs: tf.Tensor, num_filters: int, is_training: bool
    ) -> tf.Tensor:
        """Basic unit of work batch_norm->conv->dropout.

        Batch normaliazation and dropout are conditioned on the obect's HParams

        Args:
            inputs (tf.Tensor): input tensor
            num_filters (int): number of inputs for the conv operation
            is_training: indicates if the model is training

        Returns:
            tf.Tensor: the output tensor from the block operation
        """

        if self.hparams.batch_norm:
            inputs = self.batch_norm(inputs, is_training)

        inputs = self.conv(inputs, num_filters)

        if self.hparams.dropout:
            inputs = self.dropout(inputs, is_training)

        return inputs

    def batch_norm(
        self, inputs: tf.Tensor, is_training: bool
    ):  # pylint: disable=missing-docstring
        axis = 3 if self.data_format == "channels_last" else 1

        return LAYERS.batch_normalization(inputs, training=is_training, axis=axis)

    def dropout(
        self, inputs: tf.Tensor, is_training: bool
    ):  # pylint: disable=missing-docstring
        rate = self.hparams.dropout_rate if is_training else 0
        return LAYERS.dropout(inputs, rate=rate)

    def conv(
        self,
        inputs,
        num_filters,
        padding="same",
        strides=1,
        activation=tf.nn.relu,
        name="conv",
        kernel_size=3,
    ):  # pylint: disable=missing-docstring,too-many-arguments

        inputs = LAYERS.conv2d(
            inputs,
            num_filters,
            kernel_size,
            padding=padding,
            strides=strides,
            kernel_initializer=VarInit,
            use_bias=True,
            data_format=self.data_format,
            name=name,
            activation=activation,
        )

        return inputs

    def down_sample(self, inputs):
        """Reduces inputs width and height by half.

        Args:
            inputs (tf.Tensor): input tensor

        Returns:
            input tensor downsampled
        """
        pool_size = 2
        stride = 2
        return LAYERS.max_pooling2d(
            inputs, pool_size, stride, data_format=self.data_format, name="downsampler"
        )

    def up_sample(self, inputs):
        """Doubles inputs width and height.

        Transposes the input if necessary for tf.image.resize_images

        Args:
            inputs (tf.Tensor): input tensor

        Returns:
            input tensor upsampled
        """

        def wrap_tranpose(up_func, _inputs):
            _inputs = tf.transpose(_inputs, [0, 2, 3, 1])
            _inputs = up_func(_inputs)
            _inputs = tf.transpose(_inputs, [0, 3, 1, 2])

            return _inputs

        def upsample_func(_inputs):
            _, width, height, _ = _inputs.shape.as_list()

            _inputs = tf.image.resize_images(
                _inputs,
                (width * 2, height * 2),
                method=tf.image.ResizeMethod.BICUBIC,
                align_corners=True,
            )

            return _inputs

        if self.data_format == "channels_first":
            return wrap_tranpose(upsample_func, inputs)

        return upsample_func(inputs)
